unit MainFrm;

interface

uses
  Windows, SysUtils, Classes, Graphics, Clipbrd, Messages,
  Controls, Forms, Dialogs, ComCtrls, RichViewActions, RVTypes, RVTable,
  Actions, ActnList, ToolWin, RVStyle, RVScroll, RichView, RVEdit,
  CRVData, RVERVData, RVMarker, RVItem, RVRVData, SclRView;

type
  TMainForm = class(TForm)
    RVStyle1: TRVStyle;
    ToolBar1: TToolBar;
    ActionList1: TActionList;
    rvActionFontEx1: TrvActionFontEx;
    rvActionCut1: TrvActionCut;
    rvActionCopy1: TrvActionCopy;
    rvActionPaste1: TrvActionPaste;
    rvActionUndo1: TrvActionUndo;
    rvActionRedo1: TrvActionRedo;
    rvActionSelectAll1: TrvActionSelectAll;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    SRichViewEdit1: TSRichViewEdit;
    procedure FormCreate(Sender: TObject);
    procedure ToolButton8Click(Sender: TObject);
    procedure SRichViewEdit1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SRichViewEdit1Paste(Sender: TCustomRichViewEdit;
      var DoDefault: Boolean);
  private
    { Private declarations }
    procedure ApplyDeleteFont;
    procedure ApplyInsertFont;
    procedure DoDeleteMarkedText(RVData: TCustomRVData);
    procedure DoAcceptMarkedText(RVData: TCustomRVData);
    function GetItemColor(RVData: TCustomRVData; ItemNo: Integer): TColor;
    function IsDeletedText(RVData: TCustomRVData; ItemNo: Integer): Boolean;
    function IsInsertedText(RVData: TCustomRVData; ItemNo: Integer): Boolean;
    function GetNormalTextStyleNo(ARVStyle: TRVStyle;
      StyleNo: Integer): Integer;
    function IsOnlyInsertedTextSelected(rve: TCustomRichViewEdit): Boolean;

  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

const
  INSERT_COLOR = clGreen;
  DELETE_COLOR = clRed;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  // disabling drag and drop
  SRichViewEdit1.AcceptDragDropFormats := [];
end;
{------------------------------------------------------------------------------}
// Applying properties of "deleted text" to the selection
// We use TrvActionFontEx. Without RichViewActions, it can be implemented using
// ApplyStyleConversion method
procedure TMainForm.ApplyDeleteFont;
begin
  rvActionFontEx1.UserInterface := False;
  rvActionFontEx1.Font.Style := [fsStrikeOut];
  rvActionFontEx1.Font.Color := DELETE_COLOR;
  rvActionFontEx1.ValidProperties := [rvfimColor, rvfimStrikeout, rvfimUnderline];
  rvActionFontEx1.ExecuteTarget(SRichViewEdit1);
  rvActionFontEx1.UserInterface := True;
end;
{------------------------------------------------------------------------------}
// Applying properties of "inserted text" to the selection
procedure TMainForm.ApplyInsertFont;
begin
  rvActionFontEx1.UserInterface := False;
  rvActionFontEx1.Font.Style := [fsUnderline];
  rvActionFontEx1.Font.Color := INSERT_COLOR;
  rvActionFontEx1.ValidProperties := [rvfimColor, rvfimStrikeout, rvfimUnderline];
  rvActionFontEx1.ExecuteTarget(TCustomRichViewEdit(SRichViewEdit1));
  rvActionFontEx1.UserInterface := True;
end;
{------------------------------------------------------------------------------}
// Returning index of text style having all properties of RVStyle.TextStyles[StyleNo],
// but with removed properties of "inserted text"
function TMainForm.GetNormalTextStyleNo(ARVStyle: TRVStyle; StyleNo: Integer): Integer;
var
  TextStyle: TFontInfo;
begin
  TextStyle := TFontInfo.Create(nil);
  TextStyle.Assign(ARVStyle.TextStyles[StyleNo]);
  TextStyle.Color := clWindowText;
  TextStyle.Style := TextStyle.Style - [fsUnderline, fsStrikeOut];
  Result := ARVStyle.FindTextStyle(TextStyle);
  TextStyle.Free;
end;
{------------------------------------------------------------------------------}
// Returns a range of selected items.
procedure GetSelectionBounds(rve: TCustomRichViewEdit;
  out StartNo, EndNo: Integer);
var
  StartOffs, EndOffs: Integer;
begin
  rve.RVData.GetSelectionBoundsEx(StartNo, StartOffs, EndNo, EndOffs, True);
  if (StartNo <> EndNo) or (StartOffs <> EndOffs) then
  begin
    if StartOffs >= rve.GetOffsAfterItem(StartNo) then
      inc(StartNo);
    if EndOffs <= rve.GetOffsBeforeItem(EndNo) then
      dec(EndNo);
  end;
end;
{------------------------------------------------------------------------------}
// Is only "inserted text" selected?
function TMainForm.IsOnlyInsertedTextSelected(rve: TCustomRichViewEdit): Boolean;
var
  i, StartNo, EndNo: Integer;
begin
  rve := rve.TopLevelEditor;
  GetSelectionBounds(rve, StartNo, EndNo);
  Result := False;
  for i := StartNo to EndNo do
    if not IsInsertedText(rve.RVData, i) then
      exit;
  Result := True;
end;
{------------------------------------------------------------------------------}
procedure TMainForm.SRichViewEdit1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);

  procedure Deselect(RVData: TRVEditRVData);
  var
    ItemNo1, Offs1, ItemNo2, Offs2: Integer;
  begin
    RVData.GetSelectionBounds(ItemNo1, Offs1, ItemNo2, Offs2, False);
    RVData.SetSelectionBounds(ItemNo2, Offs2, ItemNo2, Offs2);
    RVData.Invalidate;
  end;

  procedure ApplyDelete(Left: Boolean);
  var
    rve: TCustomRichViewEdit;
    RVData: TRVEditRVData;
  begin
    rve := TSRichViewEdit(Sender).ActiveEditor.TopLevelEditor;
    RVData := rve.RVData as TRVEditRVData;
    TSRichViewEdit(Sender).CanUpdate := False;
    try
      if not RVData.SelectionExists(False, True) then
      begin
        RVData.StartShiftMoving;
        if Left then
          RVData.OnLeftPress(True, False)
        else
          RVData.OnRightPress(True, False);
        if RVData <> TCustomRichViewEdit(Sender).TopLevelEditor.RVData then
              exit;
      end;
      if IsOnlyInsertedTextSelected(rve) then
        rve.DeleteSelection
      else
        ApplyDeleteFont;
      Deselect(RVData);
    finally
      TSRichViewEdit(Sender).CanUpdate := True;
    end;
  end;

begin
  case Key of
    // backspace
    VK_BACK:
      begin
        ApplyDelete(True);
        Key := 0;
      end;
    // delete
    VK_DELETE:
      begin
        ApplyDelete(False);
        Key := 0;
      end;
    // keys that do not trigger insertion
    VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN, VK_HOME, VK_END, VK_NEXT, VK_PRIOR,
    VK_SHIFT, VK_CONTROL, VK_CAPITAL, VK_MENU, VK_ESCAPE, VK_INSERT, VK_F1..VK_F24: ;
    else
    // before insertion
    begin
      if TSRichViewEdit(Sender).ActiveEditor.SelectionExists then
        SendMessage(TSRichViewEdit(Sender).ActiveEditor.TopLevelEditor.Handle,
          WM_KEYDOWN, VK_DELETE, 0);
      ApplyInsertFont;
    end;
  end;
end;
{------------------------------------------------------------------------------}
// Allowing pasting only a plain text, and colorizing it properly
procedure TMainForm.SRichViewEdit1Paste(Sender: TCustomRichViewEdit;
  var DoDefault: Boolean);
var
  S: String;
  rve: TCustomRichViewEdit;
begin
  DoDefault := False;
  if Clipboard.HasFormat(CF_UNICODETEXT) then
  begin
    S := Clipboard.AsText;
    rve := Sender.TopLevelEditor;
    rve.BeginUndoGroup(rvutInsert);
    rve.SetUndoGroupMode(True);
    try
      if rve.SelectionExists then
        SendMessage(rve.Handle, WM_KEYDOWN, VK_DELETE, 0);
      ApplyInsertFont;
      rve.InsertText(S);
    finally
      rve.SetUndoGroupMode(False);
    end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TMainForm.ToolButton8Click(Sender: TObject);

  {............................................................................}
  procedure ProcessDocument(RVData: TCustomRVData);
  begin
    DoAcceptMarkedText(RVData);
    DoDeleteMarkedText(RVData);
    RVData.DeleteUnusedStyles(True, True, True);
    RVData.NormalizeDocument(0, True);
  end;
  {............................................................................}
var
  i: TSRVHeaderFooterType;
begin
  SRichViewEdit1.StartEditing(srvrveMain);
  ProcessDocument(SRichViewEdit1.RichViewEdit.RVData);
  for i := Low(TSRVHeaderFooterType) to High(TSRVHeaderFooterType) do
    ProcessDocument(SRichViewEdit1.SubDocuments[i]);
  SRichViewEdit1.RichViewEdit.ClearUndo;
  SRichViewEdit1.Format;
end;
{------------------------------------------------------------------------------}
// Returning color for text and tabs, and clNone for other items
function TMainForm.GetItemColor(RVData: TCustomRVData; ItemNo: Integer): TColor;
var
  StyleNo: Integer;
begin
  StyleNo := RVData.GetItemStyle(ItemNo);
  if (StyleNo < 0) then
    StyleNo := RVData.GetItem(ItemNo).AssociatedTextStyleNo;
  if StyleNo >= 0 then
    Result := RVData.GetRVStyle.TextStyles[StyleNo].Color
  else
    Result := clNone;
end;
{------------------------------------------------------------------------------}
function TMainForm.IsDeletedText(RVData: TCustomRVData; ItemNo: Integer): Boolean;
begin
  Result := GetItemColor(RVData, ItemNo) = DELETE_COLOR;
end;
{------------------------------------------------------------------------------}
function TMainForm.IsInsertedText(RVData: TCustomRVData; ItemNo: Integer): Boolean;
begin
  Result := GetItemColor(RVData, ItemNo) = INSERT_COLOR;
end;
{------------------------------------------------------------------------------}
// Deleting "deleted text"
procedure TMainForm.DoDeleteMarkedText(RVData: TCustomRVData);
var
  i,r,c: Integer;
  Table: TRVTableItemInfo;
  DeleteCount: Integer;
  {...............................................................}
  // Delete Count items starting from the Index-th item
  procedure DeleteText(Index, Count: Integer);
  var
    PS, BR, PB: Boolean;
    i, ParaNo: Integer;
  begin
    // storing properties of the first item that will be deleted
    PS := RVData.IsParaStart(Index);
    BR := RVData.IsFromNewLine(Index) and not PS;
    PB := RVData.PageBreaksBeforeItems[Index];
    ParaNo := RVData.GetItemPara(Index);
    // deleting
    RVData.DeleteItems(Index, Count);
    // applying the stored properties to the next item after the deleted items
    if (Index < RVData.ItemCount) and not RVData.IsParaStart(Index) then
    begin
      if PS then
        RVData.GetItem(Index).SameAsPrev := False;
      if BR then
        RVData.GetItem(Index).BR := True;
      if PB then
        RVData.PageBreaksBeforeItems[Index] := True;
      for i := Index to RVData.ItemCount - 1 do
        RVData.GetItem(i).ParaNo := ParaNo;
    end;
  end;
  {...............................................................}
begin
  DeleteCount := 0;
  for i := RVData.ItemCount - 1 downto 0 do
  begin
    if IsDeletedText(RVData, i) then
      inc(DeleteCount)
    else if DeleteCount > 0 then
    begin
      DeleteText(i+1, DeleteCount);
      DeleteCount := 0;
    end;
    if RVData.GetItem(i) is TRVTableItemInfo then
    begin
      Table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to Table.RowCount - 1 do
        for c := 0 to Table.ColCount - 1 do
          if Table.Cells[r,c] <> nil then
            DoDeleteMarkedText(Table.Cells[r,c].GetRVData);
    end;
  end;
  if DeleteCount > 0 then
    DeleteText(0, DeleteCount);
end;
{------------------------------------------------------------------------------}
// Accepting "inserted text"
// We could simply change properties of TextStyles[] in TRVStyle,
// but it can lead to duplicate styles with identical properties.
// So we do it per-item
procedure TMainForm.DoAcceptMarkedText(RVData: TCustomRVData);

  function GetItemTextStyle(ItemNo: Integer): Integer;
  begin
    Result := RVData.GetItemStyle(ItemNo);
    if Result < 0 then
      Result := RVData.GetItem(ItemNo).AssociatedTextStyleNo;
  end;

  procedure SetItemTextStyle(ItemNo, StyleNo: Integer);
  begin
    if RVData.GetItemStyle(ItemNo) < 0 then
      RVData.GetItem(ItemNo).AssociatedTextStyleNo := StyleNo
    else
      RVData.GetItem(ItemNo).StyleNo := StyleNo
  end;

var
  i,r,c: Integer;
  Table: TRVTableItemInfo;
begin
  for i := 0 to RVData.ItemCount - 1 do
    if IsInsertedText(RVData, i) then
      SetItemTextStyle(i, GetNormalTextStyleNo(RVData.GetRVStyle, GetItemTextStyle(i)))
    else if RVData.GetItem(i) is TRVTableItemInfo then
    begin
      Table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to Table.RowCount - 1 do
        for c := 0 to Table.ColCount - 1 do
          if Table.Cells[r,c] <> nil then
            DoAcceptMarkedText(Table.Cells[r,c].GetRVData);
    end;
end;


end.
